/**
 * PortalManager class, implementation of an API for communication and interaction
 * between portlets hosted inside the NexJ mobile application.
 */
function PortalManager()
{
    // Context variable values, structure: <String workspaceId, Map workspaceContextMap>.
    // workspaceContextMap structure: <String variableName, String variableValue>.
    this.contextMap = {}
    
    // Context variable listeners, structure: <String workspaceId, Map workspaceListenerMap>
    // workspaceListenerMap structure: <String viewId, Array subscribedVariables>
    // The workspaceListenerMap is indexed by viewId to allow grouping of deltas when firing sysContext
    // and for ease of unregistering views
    this.listenerMap = {}
    
    // Views to workspaces, structure: <String viewId, String workspaceId>
    this.viewToWorkspaceMap = {}
    
    // Event handlers for each view, structure: <String viewId, String eventHandlerFunctionName>
    this.eventHandlerMap = {}
    
    //Views to PortletGroup indexs, structure: <String viewId, int portletGroupIndex>
    this.viewToPortletGroupMap = {}
    
    this.appConfigJson = null
    
    // Array of portlet groups
    this.portletGroups = null
}

window.portalManager = new PortalManager();

// Functions which can be invoked with the public api dispatcher
PortalManager.PUBLIC_API =
{
    getContext:true,
    getWorkspaceId:true,
    registerEventHandler:true,
    removeContextListener:true,
    setBusy:true,
    setContext:true
}

// Public API
PortalManager.prototype.invokeAPI = PortalManager_invokeAPI

PortalManager.prototype.getContext = PortalManager_getContext
PortalManager.prototype.getWorkspaceId = PortalManager_getWorkspaceId
PortalManager.prototype.registerEventHandler = PortalManager_registerEventHandler
PortalManager.prototype.removeContextListener = PortalManager_removeContextListener
PortalManager.prototype.setBusy = PortalManager_setBusy
PortalManager.prototype.setContext = PortalManager_setContext

// Internal API
PortalManager.prototype.invokeAPIInternal = PortalManager_invokeAPIInternal

PortalManager.prototype.addContextListener = PortalManager_addContextListener
PortalManager.prototype.addWorkspace = PortalManager_addWorkspace
PortalManager.prototype.fireEvent = PortalManager_fireEvent
PortalManager.prototype.fireEventInternal = PortalManager_fireEventInternal
PortalManager.prototype.generateGUID = PortalManager_generateGUID
PortalManager.prototype.getSimpleType = PortalManager_getSimpleType
PortalManager.prototype.loadApplication = PortalManager_loadApplication
PortalManager.prototype.notifyContextListeners = PortalManager_notifyContextListeners
PortalManager.prototype.processLayoutJson = PortalManager_processLayoutJson
PortalManager.prototype.processWorkspaceJson = PortalManager_processWorkspaceJson

PortalManager.prototype.dumpMaps = PortalManager_dumpMaps

/****************************************************************************************
 * Public API Implementations
 * These functions can be called from PortalAPI via the phonegap plugin.
 */

/**
 * Public API dispatcher, invoked by native code for PortalAPI calls.
 *
 * @param sCallerViewId ViewId of the caller
 * @param sSelector Name of the function to invoke.
 * @param args Array of arguments to pass to the selected function.
 */
function PortalManager_invokeAPI(sCallerViewId, sSelector, args)
{
    console.log('PM Invoke API: ' + JSON.stringify(arguments))
    
    // Only allow public api methods to be called
    if (PortalManager.PUBLIC_API[sSelector] && PortalManager.prototype[sSelector])
    {
        /*
         Internally, the first argument to all public API functions is the id of the calling view.
         This argument is added automatically by the phonegap plugin.
         */
        args.splice(0, 0, sCallerViewId);

        return PortalManager.prototype[sSelector].apply(this, args);
    }
    else
    {
        return null;
    }
}

/**
 * Get the value of a context variable.
 * 
 * @param sCallerViewId ViewId of the caller
 * @param sVarName Name of the context variable
 * @return Value of the context variable
 */
function PortalManager_getContext(sCallerViewId, sVarName)
{
    var wkspId = this.getWorkspaceId(sCallerViewId);
    
    // Only allow retrieval of subscribed context variables
    if (wkspId && this.listenerMap[wkspId][sCallerViewId] &&
        (this.listenerMap[wkspId][sCallerViewId].indexOf(sVarName) != -1))
    {
        return this.contextMap[wkspId][sVarName];
    }
    else
    {
        return null;
    }
}

/**
 * Stub function for implementation of PortalAPI.setBusy
 */
function PortalManager_setBusy(sCallerViewId, portletId)
{
    return true;
}

/**
 * Set context variables in the workspace scope of the calling view
 *
 * @param sCallerViewId Id of the portlet setting context variables.
 * @param newContextValues Map of new context values, indexed by context variable name.
 */
function PortalManager_setContext(sCallerViewId, newContextValues)
{
    console.log('PM setContext: ' + JSON.stringify(arguments))
    
    var wkspId = this.getWorkspaceId(sCallerViewId);

    if (!this.contextMap[wkspId])
    {
        this.contextMap[wkspId] = {};
    }

    var deltaMap = {}
    
    /*
     * Build the delta map, <String variableName, DeltaObject { oldValue, newValue }>
     * Only add entries if the value has actually changed.
     */
    for (var sVarName in newContextValues)
    { 
        var sVarValue = newContextValues[sVarName]
        
        if (sVarName)
        {
            var sOldValue = this.contextMap[wkspId][sVarName]
            
            if (sOldValue != sVarValue)
            {
                this.contextMap[wkspId][sVarName] = sVarValue
                
                deltaMap[sVarName] = { 'oldValue' : sOldValue, 'newValue' : sVarValue }
            }
        }
        
        // If any values have changed, notify listeners
        if (Object.keys(deltaMap).length > 0)
        {
            this.notifyContextListeners(sCallerViewId, deltaMap)
        }
    }
    
    return true
}

/**
 * Remove all variable listeners for the calling view 
 *
 * @param sCallerViewId id of the calling view.
 */
function PortalManager_removeContextListener(sCallerViewId)
{
    var wkspId = this.getWorkspaceId(sCallerViewId);

    if (wkspId)
    {
        this.listenerMap[wkspId][sCallerViewId] = null
    }
    
    return true
}

/**
 * Register an event handler for the calling view.
 *
 * @param sCallerViewId id of the calling view.
 * @param sHandlerFunctionName name of the event handler function.
 */
function PortalManager_registerEventHandler(sCallerViewId, sHandlerFunctionName)
{
    // TODO: security check on handler function name?
    this.eventHandlerMap[sCallerViewId] = sHandlerFunctionName
    
    return true
}

/****************************************************************************************
 * Internal API Implementations
 * These functions should never be invoked by the PortalAPI or through the phonegap plugin.
 */

/**
 * Dispatch function for the internal API functions.
 * @private
 *
 * @param sSelector Name of the function to invoke.
 * @param args Array of arguments to pass to the selected function.
 */
function PortalManager_invokeAPIInternal(sSelector, args)
{
    console.log('PM InvokeAPIInternal: ' + JSON.stringify(arguments))
    if (PortalManager.prototype[sSelector])
    {
        return PortalManager.prototype[sSelector].apply(this, args);
    }
    else
    {
        return null;
    }
}

/**
 * Subscribe a view to context variables.
 *
 * @param sViewId id of the view to subscribe.
 * @param varArray array of variables the view is subscribing to.
 */
function PortalManager_addContextListener(sViewId, varArray)
{
    var wkspId = this.getWorkspaceId(sViewId);
    
    if (wkspId)
    {
        if (this.listenerMap[wkspId] == null)
        {
            this.listenerMap[wkspId] = varArray
        }
        else
        {
            for (var i = 0; i < varArray.length; i++)
            {
                this.listenerMap[wkspId].push(varArray[i])
            }
        }
    }
    
    return true
}

/**
 * Initialize maps for a workspace
 * @private
 */
function PortalManager_addWorkspace(sWorkspaceId)
{
    this.contextMap[sWorkspaceId] = {}
    this.listenerMap[sWorkspaceId] = {}
}

/**
 * Debug function to dump the state of the Portal Manager
 */
function PortalManager_dumpState()
{
    var state = JSON.stringify(this)
    console.log(state)
    
    return state
}

/**
 * Fire an event to a view.
 * @TODO: expose this as public api
 */
function PortalManager_fireEvent(sEventName, params, sTargetViewId)
{
    var oEvent = {}
    
    oEvent.name = sEventName
    oEvent.parameter = params
    
    this.fireEventInternal([oEvent], sTargetViewId)
}

/**
 * Internal function for firing multiple events to a view.
 * @private
 */
function PortalManager_fireEventInternal(events, sTargetViewId)
{
    console.log('PM fireEventInternal: ' + JSON.stringify(arguments))
    // TODO: security check on params?
    var sHandlerFunctionName = this.eventHandlerMap[sTargetViewId]
    
    console.log('PM fireEventInternal handlerFunctionName: ' + sHandlerFunctionName)
    
    if (sHandlerFunctionName)
    {
        window.plugins.portal.execJS(sTargetViewId, sHandlerFunctionName + '(' + JSON.stringify(events) + ')')
    }
}

/**
 * Function to create a random 128bit guid.
 * @return Random GUID as a hex string (Not a binary number, leading zeros are allowed)
 */
function PortalManager_generateGUID()
{
    var symbols = '0123456789ABCDEF'.split('')
    var guid = new Array(32);
    var rand = 0;
    
    for (var i = 0; i < 32; i++)
    {
        /* 
           Minimize calls to Math.random() by using the random
           number bitwise. We use the last 32 bits so the 33rd bit
           is set to 1 as a sentinel.
         */
        if (rand < 0x2)
        {
            rand = ((Math.random()*0x100000000) & 0xffffffff) | 0x100000000
        }
        
        guid[i] = symbols[rand & 0xf]
        
        rand = rand >> 4
    }
    
    return guid.join('')
}

/**
 * Parse the $type string from the workspace layout.
 * @private
 
 * @param $type string
 * @return The simple type string. Null if sType is invalid.
 */
function PortalManager_getSimpleType(sType)
{
    if (sType) {
        var index = sType.lastIndexOf(':')
        
        if (index != -1) {
            return sType.substring(index+1)
        }
    }
    
    return null
}

/**
 * Get the id of the containing workspace for a view
 * @private
 */
function PortalManager_getWorkspaceId(sViewId)
{
    return this.viewToWorkspaceMap[sViewId]
}

/**
 * Load the application based on the supplied configuration.
 * @private
 *
 * @param jsonConfig JSON object with the mobile portal configuration
 */
function PortalManager_loadApplication(data)
{
    this.appConfigJson = data
    
    if (data)
    {
        var workspaces = this.workspaceJson = data.workspaces
        var portletGroups = this.portletGroupJson = data.portletGroups
        
        if (Array.isArray(portletGroups))
        {
            this.portletGroups = portletGroups
        }
        
        if (Array.isArray(workspaces))
        {
            for (var i = 0; i < workspaces.length; i++)
            {
                var viewId = this.generateGUID()
                workspaces[i].viewId = viewId
                
                this.addWorkspace(viewId)
                this.processWorkspaceJson(workspaces[i])

                window.plugins.portal.createLayoutView(null, workspaces[i])
            }
            
            window.plugins.portal.setDefaultWorkspace()
        }
    }
    
    // TODO: store workspaces with timestamps
}

/**
 * Parse the JSON for a workspace and setup the 
 * structures in the PortalManager.
 * @private
 */
function PortalManager_processWorkspaceJson(workspaceJson)
{
    if (this.getSimpleType(workspaceJson['$type']) == 'Single')
    {
        var workspaceId = workspaceJson.viewId
        this.viewToWorkspaceMap[workspaceId] = workspaceId
        
        this.processLayoutJson(workspaceId, workspaceJson)
    }
    
    // TODO: Handle other types of layouts, Tab, CWP app perspective etc.
}

/**
 * Parse the JSON for a layout and set the layout state.
 * @private
 */
function PortalManager_processLayoutJson(workspaceId, layoutJson)
{
    if (layoutJson.context)
    {
        this.listenerMap[workspaceId][layoutJson.viewId] = layoutJson.context;
    }
    
    if (layoutJson.portletGroupIndex) {
        this.viewToPortletGroupMap[layoutJson.viewId] = layoutJson.portletGroupIndex
    }
    
    // TODO: Handle other types of layouts, Tab, CWP app perspective etc.
}

/**
 * Notify the context listeners when context variables change.
 * @private
 * 
 * @param wkspId Id of the workspace where the variables changed.
 * @param deltaMap DeltaMap of the changed variables.
 */
function PortalManager_notifyContextListeners(wkspId, deltaMap)
{
    console.log('PM notifyContextListeners: ' + JSON.stringify(deltaMap))
    
    var wkspListenerMap = this.listenerMap[wkspId]
    var viewIdList = Object.keys(wkspListenerMap)
    
    /*
     * Iterate through the listener map and for each subscribed variable, 
     * if there is change in value, fire sysContext on the listening portlet.
     */
    for (var i = 0; i < viewIdList.length; i++)
    {
        var viewId = viewIdList[i]
        var subscribedVarList = wkspListenerMap[viewId]
        var viewDelta = {}
        
        if (subscribedVarList)
        {
            for (var j = 0; j < subscribedVarList.length; j++)
            {
                var sVarName = subscribedVarList[i]
                var delta = deltaMap[sVarName]

                if (delta)
                {
                    viewDelta[sVarName] = { 'oldValue' : delta.oldValue, 'newValue' : delta.newValue }
                }
            }
            
            if (Object.keys(viewDelta).length > 0)
            {
                this.fireEvent('sysContext', [viewDelta], viewId)
            }
        }
    }
}
